###############################################################################
#
#    NAME: wkcffreesurfaceflow.tcl
#
#    PURPOSE: Write free surface flow application data file
#
#    QUANTECH ATZ-DEVELOPMENT DEPARTMENT
#
#    AUTHOR : G. Socorro => GSM
#
#    CREATED AT: 30/05/14
#
#    HISTORY:
#
#     0.3- 11/07/14-GSM, add the proc WriteFreeSurfaceFlowDistanceNodalData
#     0.2- 07/07/14-GSM, modify the proc WriteFreeSurfaceFlowMaterialProperties to get the material data from the combobox Fluid Material => Material
#     0.1- 30/05/14-GSM, create a base source code from wkcf.tcl
#
###############################################################################

proc ::wkcf::WriteFreeSurfaceFlowDistanceNodalData {AppId} {
    # ABSTRACT : Write distance to all fluid elements nodes
    variable filechannel
  
    set ditem "DISTANCE"
    set dval "-1.0"
    set cpropid "1"
       
    # Get all defined element types
    set fluidelemlist [list "LevelSetFluid2D" "LevelSetFluid3D"]
    set cxpath "$AppId//c.Elements"
    set glist [::xmlutils::setXmlContainerIds $cxpath]
    set kwxpath "Applications/$AppId"
    # WarnWinText "glist:$glist"
    foreach celemid $glist {
	# Get the group identifier defined for this element 
	set cxpath "$AppId//c.Elements//c.[list ${celemid}]"
	set cgrouplist [::xmlutils::setXmlContainerIds $cxpath]
	if {([llength $cgrouplist]) && ($celemid in $fluidelemlist)} { 
	    # WarnWinText "cgrouplist:$cgrouplist"
	    foreach cgroupid $cgrouplist {
		# Check if this group is in active state
		# Get active property
		set cxpath "$AppId//c.Elements//c.[list ${celemid}]//c.[list ${cgroupid}]"
		set cproperty "active"
		set ActiveGroup [::xmlutils::setXml $cxpath $cproperty]
		# WarnWinText "ActiveGroup:$ActiveGroup"
		if {$ActiveGroup =="0"} {
		  continue
		}
		if { [GiD_EntitiesGroups get $cgroupid nodes -count] } {
		    GiD_File fprintf $filechannel "%s" "Begin NodalData $ditem \/\/ Distance condition GUI fluid group identifier: $cgroupid"
		    foreach node_id [GiD_EntitiesGroups get $cgroupid nodes] {
			GiD_File fprintf $filechannel "%10i %8i  %10.5e" $node_id $cpropid $dval
		    }
		    GiD_File fprintf $filechannel "%s" "End NodalData"
		    GiD_File fprintf $filechannel "%s" ""
		}
	    }
	}
    }
}

proc ::wkcf::WriteFreeSurfaceFlowInletVelocityBC {AppId inletvelglist kwordlist} {
    # ABSTRACT: Write the free surface boundary conditions
    variable ndime; variable dprops; variable filechannel

    # WarnWinText "inletvelglist:$inletvelglist\nkwordlist:$kwordlist"
   
    if {([llength $inletvelglist])} {
	# For debug
	if {!$::wkcf::pflag} {
	    set inittime [clock seconds]
	}

	set icondid "InletVelocity"
	set cpropid "1"
	set xitem [lindex $kwordlist 0]
	set yitem [lindex $kwordlist 1]
	set zitem [lindex $kwordlist 2]
	
	# Write the inlet boundary condition properties
    
	# For each group in the inlet condition
	foreach igroupid $inletvelglist {
	    lassign $dprops($AppId,BC,$icondid,$igroupid,GProps) ix ixval iy iyval iz izval
	    # wa "ix:$ix ixval:$ixval iy:$iy iyval:$iyval iz:$iz izval:$izval"
	    # X component
	    if {$ix} {
		if { [GiD_EntitiesGroups get $igroupid nodes -count] } {
		    GiD_File fprintf $filechannel "%s" "Begin NodalData $xitem \/\/ Inlet velocity condition GUI group identifier: $igroupid"
		    foreach node_id [GiD_EntitiesGroups get $igroupid nodes] {
		        GiD_File fprintf $filechannel "%10i %8i %10.5e" $node_id $cpropid $ixval
		    }
		    GiD_File fprintf $filechannel "%s" "End NodalData"
		    GiD_File fprintf $filechannel "%s" ""
		}
	    }
	    
	    # Y component
	    if {$iy} {
		if { [GiD_EntitiesGroups get $igroupid nodes -count] } {
		    GiD_File fprintf $filechannel "%s" "Begin NodalData $yitem \/\/ Inlet velocity condition GUI group identifier: $igroupid"
		    foreach node_id [GiD_EntitiesGroups get $igroupid nodes] {
		        GiD_File fprintf $filechannel "%10i %8i %10.5e" $node_id $cpropid $iyval
		    }
		    GiD_File fprintf $filechannel "%s" "End NodalData"
		    GiD_File fprintf $filechannel "%s" ""
		}
	    }
	    
	    # Z component
	    if {$ndime =="2D"} {
		set iz 1 
	    }
	    if {$iz} {
		if { [GiD_EntitiesGroups get $igroupid nodes -count] } {
		    GiD_File fprintf $filechannel "%s" "Begin NodalData $zitem \/\/ Inlet velocity condition GUI group identifier: $igroupid"
		    foreach node_id [GiD_EntitiesGroups get $igroupid nodes] {
			GiD_File fprintf $filechannel "%10i %8i %10.5e" $node_id $cpropid $izval
		    }
		    GiD_File fprintf $filechannel "%s" "End NodalData"
		    GiD_File fprintf $filechannel "%s" ""
		}
	    }
       	}
    
       # Write distance for all nodes of the inlet BC
       set ditem "DISTANCE"
       set dval "-1.0"
       # For each group in the inlet condition
	foreach igroupid $inletvelglist {
	    if { [GiD_EntitiesGroups get $igroupid nodes -count] } {
		GiD_File fprintf $filechannel "%s" "Begin NodalData $ditem \/\/ Distance condition GUI group identifier: $igroupid"
		foreach node_id [GiD_EntitiesGroups get $igroupid nodes] {
		    GiD_File fprintf $filechannel "%10i %8i  %10.5e" $node_id $cpropid $dval
		}
		GiD_File fprintf $filechannel "%s" "End NodalData"
		GiD_File fprintf $filechannel "%s" ""
	    }
	}
	
	# For debug
	if {!$::wkcf::pflag} {
	    set endtime [clock seconds]
	    set ttime [expr {$endtime-$inittime}]
	    # WarnWinText "endtime:$endtime ttime:$ttime"
	    WarnWinText "Write fluid inlet boundary conditions: [::KUtils::Duration $ttime]"
	}
    }
}

proc ::wkcf::WriteFreeSurfaceFlowMaterialProperties {AppId fileid} {
    
    set rootdataid $AppId
  
    set getmatpmethod 1
    
    if {$getmatpmethod==1} {
	set cproperty "dv"
	# Get assigned material identifier
	set cxpath "$rootdataid//c.FSF-Material//i.Material"
	set SelectedMaterialId [::xmlutils::setXml $cxpath $cproperty]
	# wa "SelectedMaterialId:$SelectedMaterialId"
	if {$SelectedMaterialId !=""} {
	    set mpxpath "[::KMat::findMaterialParent $SelectedMaterialId]//m.[list ${SelectedMaterialId}]"
	    # WarnWinText "mpxpath:$mpxpath"
	    # Get the material properties
	    set pid "Density"
	    set xpath "c.General"
	    set Density [lindex [::KMat::getMaterialProperties "p" "$mpxpath//$xpath//p.[list $pid]"] 0 1]
	    set pid "Viscosity"
	    set xpath "c.Fluid"
	    set Viscosity [lindex [::KMat::getMaterialProperties "p" "$mpxpath//$xpath//p.[list $pid]"] 0 1]
	    # wa "Density:$Density Viscosity:$Viscosity"
	    puts $fileid "Density = [format {%g} $Density]"
	    puts $fileid "Viscosity = [format {%g} $Viscosity]"
	}
    } elseif {$getmatpmethod==2} {
	# Check for use fluid element
	set usefluidelem "0"
       
       set fluidelemlist [list "LevelSetFluid2D" "LevelSetFluid3D"]
       
       # Get all defined element types
       set cxpath "$rootdataid//c.Elements"
       set glist [::xmlutils::setXmlContainerIds $cxpath]
       set kwxpath "Applications/$rootdataid"
       # WarnWinText "glist:$glist"
       foreach celemid $glist {
	   # Get the group identifier defined for this element 
	   set cxpath "$rootdataid//c.Elements//c.[list ${celemid}]"
	   set cgrouplist [::xmlutils::setXmlContainerIds $cxpath]
	   if {[llength $cgrouplist]>0} { 
	       # WarnWinText "cgrouplist:$cgrouplist"
	       foreach cgroupid $cgrouplist {
		   # Check if this group is in active state
		   # Get active property
		   set cxpath "$rootdataid//c.Elements//c.[list ${celemid}]//c.[list ${cgroupid}]"
		   set cproperty "active"
		   set ActiveGroup [::xmlutils::setXml $cxpath $cproperty]
		   # WarnWinText "ActiveGroup:$ActiveGroup"
		   if {$ActiveGroup =="0"} {
		       continue
		   } else {
		       if {$celemid in $fluidelemlist} {
			   set usefluidelem 1
			   break
		       }
		   }
	       }
	   }
       }
       
       # Write material properties
       if {$usefluidelem} {
	   set Density [::wkcf::GetFluidMaterialProperties $AppId "PropertyId" "Density"]
	   set Viscosity [::wkcf::GetFluidMaterialProperties $AppId "PropertyId" "Viscosity"]
	   # wa "Density:$Density Viscosity:$Viscosity"
	   puts $fileid "Density = [format {%.5f} $Density]"
	   puts $fileid "Viscosity = [format {%.5f} $Viscosity]"
       }
    }
}

proc ::wkcf::WriteFreeSurfaceFlowProjectParameters {AppId fileid PDir} {
    variable ndime; variable dprops
   
    # wa "AppId:$AppId fileid:$fileid PDir:$PDir"
   
    set rootid "$AppId"
    
    set trailing_spaces  "    "

    # Kratos key word xpath
    set kxpath "Applications/$rootid"
    
    # Get others properties
    set cproperty "dv"

    set domain_size 2
    if {$ndime =="2D"} {
	set domain_size 2
    } elseif {$ndime =="3D"} {
	set domain_size 3
    }
    
    # Domain size
    puts $fileid "domain_size = $domain_size"
    puts $fileid ""
    puts $fileid ""
  
    # Write the body force properties
    set contid "LevelSetBodyForce"
    set cxpath "$rootid//c.SolutionStrategy//c.[list ${contid}]//i.GravityValue"
    set GravityValue [::xmlutils::setXml $cxpath $cproperty]
    set cxpath "$rootid//c.SolutionStrategy//c.[list ${contid}]//i.Cx"
    set Cx [::xmlutils::setXml $cxpath $cproperty]
    set cxpath "$rootid//c.SolutionStrategy//c.[list ${contid}]//i.Cy"
    set Cy [::xmlutils::setXml $cxpath $cproperty]
    set cxpath "$rootid//c.SolutionStrategy//c.[list ${contid}]//i.Cz"
    set Cz [::xmlutils::setXml $cxpath $cproperty]
    # wa "GravityValue:$GravityValue Cx:$Cx Cy:$Cy Cz:$Cz"
    set kwordlist [list "body_force_x" "body_force_y" "body_force_z"]
    set valuelist [list [expr {$GravityValue*$Cx}] [expr {$GravityValue*$Cy}] [expr {$GravityValue*$Cz}]]
    foreach ckword $kwordlist cvalue $valuelist {
	puts $fileid "$ckword = [format {%.5f} $cvalue]"
    }
    puts $fileid ""
    puts $fileid ""
  
    # Porous zone properties
    set contid "PorousZones"
    set cxpath "$rootid//c.SolutionStrategy//c.[list ${contid}]//i.UseErgunEquation"
    set UseErgunEquation [::xmlutils::setXml $cxpath $cproperty]
    if {$UseErgunEquation eq "Yes"} {
	puts $fileid "UseErgun = True "
    } else {
	puts $fileid "UseErgun = False "
    }

    puts $fileid ""
    puts $fileid ""

    # Material properties (Density and viscosity)
    ::wkcf::WriteFreeSurfaceFlowMaterialProperties $AppId $fileid
   

    puts $fileid ""
    puts $fileid ""

    # Advanced parameters
    set adplist [list "DynamicTau" "OssSwitch" "UseMassCorrection" "EdgeDetectionAngle" "UseConstantPressure" "RedistanceFrequency" \
		 "ExtrapolationLayers" "StabdtPressureFactor" "StabdtConvectionFactor" "ReductionOnFailure" "WallRoughness"]
    set keywordlist [list "tau2_factor" "" "use_mass_correction" "edge_detection_angle" "assume_constant_pressure" "redistance_frequency" \
		     "extrapolation_layers" "stabdt_pressure_factor" "stabdt_convection_factor" "reduction_on_failure" "wall_law_y"]
    foreach varid $adplist keyid $keywordlist {
	# Set xpath
	set cxpath "$rootid//c.SolutionStrategy//c.Advanced//i.${varid}"
	set cvalue [::xmlutils::setXml $cxpath $cproperty]
	if {($cvalue !="") && ($keyid !="")} {
	    if {($varid eq "UseMassCorrection") ||($varid eq "UseConstantPressure")} {
		if {$cvalue eq "Yes"} {
		    set cvalue "True"
		} elseif {$cvalue eq "No"} {
		    set cvalue "False"
		}
	    }
	   puts $fileid "$keyid = $cvalue"
	}
    }

    puts $fileid ""
    puts $fileid ""

    # compute_porous_resistance_law = 0.00000e+00

    puts $fileid ""

    # Problem parameters
    set StartTime 0.0
    set EndTime 1.0
    
    # Start time
    set cxpath "$rootid//c.ProblemParameters//i.StartTime"
    set StartTime [::xmlutils::setXml $cxpath $cproperty]
    
    # End time
    set cxpath "$rootid//c.ProblemParameters//i.EndTime"
    set EndTime [::xmlutils::setXml $cxpath $cproperty]
   
    # WarnWinText "StartTime:$StartTime EndTime:$EndTime"
    puts $fileid "Start_time = [format {%.5f} $StartTime]"
    puts $fileid "max_time = [format {%.5f} $EndTime]"
     
    # CFL safaty factor
    set cxpath "$rootid//c.ProblemParameters//i.SafetyFactor"
    set SafetyFactor [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "safety_factor = [format {%.5f} $SafetyFactor]"
    
    # Max time step
    set cxpath "$rootid//c.ProblemParameters//i.MaxTimeStep"
    set MaxTimeStep [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "max_time_step = [format {%.5f} $MaxTimeStep]"

    # Number of initial time step     
    set cxpath "$rootid//c.ProblemParameters//i.DivergenceCleareanceStep"
    set DivergenceCleareanceStep [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "number_of_inital_steps = $DivergenceCleareanceStep" 
    
    # Initial time step
    set cxpath "$rootid//c.ProblemParameters//i.InitialTimeStep"
    set InitialTimeStep [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "initial_time_step = [format {%.5f} $InitialTimeStep]"
    
    puts $fileid ""

    puts $fileid "#output settings"
    # Output step 
    set cxpath "$rootid//c.Results//i.OutputDeltaTime"
    set OutputDeltaTime [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "output_time = [format {%.5f} $OutputDeltaTime]"
    # WarnWinText "OutputDeltaTime:$OutputDeltaTime"
    
    # GiD result file
    set cxpath "$rootid//c.Results//i.GiDMultiFileFlag"
    set GiDMultiFileFlag [::xmlutils::setXml $cxpath $cproperty] 
    puts $fileid "GiDMultiFileFlag = \"$GiDMultiFileFlag\""
    
    # Print in layers
    set cxpath "$rootid//c.Results//i.PrintInLayer"
    set PrintInLayer [::xmlutils::setXml $cxpath $cproperty]
    if {$PrintInLayer eq "Yes"} {
	puts $fileid "print_layers =  True"
    } else {
	puts $fileid "print_layers =  False"
    }
    
    puts $fileid ""
    set PName [::KUtils::GetPaths "PName"]
    puts $fileid "problem_name=\"${PName}${AppId}\"" 
    puts $fileid "problem_path=\"[file join $PDir]\"" 
    
    # Get the kratos path 
    set cxpath "GeneralApplicationData//c.ProjectConfiguration//i.KratosPath"
    set cproperty "dv"
    set KratosPath [::xmlutils::setXml $cxpath $cproperty]
    set KratosPath [file native $KratosPath]
    
    # Write the kratos path
    puts $fileid "kratos_path=\"[file join $KratosPath]\""
    
}
